#include "mapEditor.hpp"
#include "collision.hpp"
#include "imgui.h"
#include "imgui_impl_sdl2.h"
#include "imgui_impl_sdlrenderer.h"
#include "map.hpp"
#include "mapEditorState.hpp"
#include "misc/cpp/imgui_stdlib.h"
#include "phoenix.hpp"
#include "state.hpp"
#include <SDL.h>
#include <SDL_events.h>
#include <SDL_keycode.h>
#include <SDL_mouse.h>
#include <SDL_render.h>
#include <stdio.h>

#if !SDL_VERSION_ATLEAST(2, 0, 17)
#error This backend requires SDL 2.0.17+ because of SDL_RenderGeometry() function
#endif

// Main code
void runMapEditor() {
  // Setup SDL
  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_GAMECONTROLLER) !=
      0) {
    printf("Error: %s\n", SDL_GetError());
    return;
  }

  // From 2.0.18: Enable native IME.
#ifdef SDL_HINT_IME_SHOW_UI
  SDL_SetHint(SDL_HINT_IME_SHOW_UI, "1");
#endif

  // Create window with SDL_Renderer graphics context
  SDL_WindowFlags window_flags =
      (SDL_WindowFlags)(SDL_WINDOW_RESIZABLE | SDL_WINDOW_ALLOW_HIGHDPI);
  SDL_Window *window =
      SDL_CreateWindow("Phoenix level editor", SDL_WINDOWPOS_CENTERED,
                       SDL_WINDOWPOS_CENTERED, 1280, 720, window_flags);
  state.renderer = SDL_CreateRenderer(
      window, -1, SDL_RENDERER_PRESENTVSYNC | SDL_RENDERER_SOFTWARE);
  if (state.renderer == nullptr) {
    SDL_Log("Error creating SDL_Renderer!");
    return;
  }
  // SDL_RendererInfo info;
  // SDL_GetRendererInfo(renderer, &info);
  // SDL_Log("Current SDL_Renderer: %s", info.name);

  // Setup Dear ImGui context
  IMGUI_CHECKVERSION();
  ImGui::CreateContext();
  ImGuiIO &io = ImGui::GetIO();
  (void)io;
  io.ConfigFlags |=
      ImGuiConfigFlags_NavEnableKeyboard; // Enable Keyboard Controls
  io.ConfigFlags |=
      ImGuiConfigFlags_NavEnableGamepad; // Enable Gamepad Controls

  // Setup Dear ImGui style
  ImGui::StyleColorsDark();

  // Setup Platform/Renderer backends
  ImGui_ImplSDL2_InitForSDLRenderer(window, state.renderer);
  ImGui_ImplSDLRenderer_Init(state.renderer);

  ImVec4 clear_color = ImVec4(0.f, 0.f, 0.f, 1.00f);
  EditorGlobalState editorState{};

  // Load textures
  PRunBoostrap();

  // Main loop
  bool done = false;
  while (!done) {
    SDL_Event event;
    while (SDL_PollEvent(&event)) {
      ImGui_ImplSDL2_ProcessEvent(&event);

      if (editorState.placingPlayerSpawn && event.type == SDL_MOUSEBUTTONDOWN) {
        if (event.button.button == SDL_BUTTON_LEFT) {
          int mouseX;
          int mouseY;
          SDL_GetMouseState(&mouseX, &mouseY);

          SDL_Log("Player spawn is: (%i,%i)", mouseX, mouseY);

          editorState.currentMap.mapPlayerSpawnX = mouseX;
          editorState.currentMap.mapPlayerSpawnY = mouseY;

          SDL_Log("Placed player spawn point");
          editorState.placingPlayerSpawn = false;
        }
      } else {
        if (event.type == SDL_MOUSEBUTTONDOWN && !ImGui::IsAnyItemActive() &&
            !ImGui::IsWindowFocused(ImGuiFocusedFlags_AnyWindow) &&
            !ImGui::IsAnyItemHovered()) {
          if (event.button.button == SDL_BUTTON_LEFT) {
            if (editorState.placeTileMode) {
              for (int i = 0; i < editorState.tilePlaceAmount; i++) {
                editorState.currentMap.data.mapTiles.push_back(
                    editorState.selectedTileName);
              }
            }
          } else if (event.button.button == SDL_BUTTON_RIGHT) {
            if (editorState.placeTileMode) {
              for (int i = 0; i < editorState.tilePlaceAmount; i++) {
                editorState.currentMap.data.mapTiles.pop_back();
              }
            }
          }
        }
      }
      if (event.type == SDL_KEYDOWN) {
        if (event.key.keysym.scancode == SDL_SCANCODE_PAGEDOWN) {
          editorState.placeTileMode = !editorState.placeTileMode;
          SDL_Log("Toggled place tile on click");
        }
      }

      if (event.type == SDL_QUIT)
        done = true;
      if (event.type == SDL_WINDOWEVENT &&
          event.window.event == SDL_WINDOWEVENT_CLOSE &&
          event.window.windowID == SDL_GetWindowID(window))

        done = true;
    }
    // Start the Dear ImGui frame
    ImGui_ImplSDLRenderer_NewFrame();
    ImGui_ImplSDL2_NewFrame();
    ImGui::NewFrame();

    ImGui::Begin("Phoenix Map Editor");

    ImGui::Text("The Phoenix map editor outputs .map files which can be loaded "
                "by phoenix");
    ImGui::InputText("##Level file path", &editorState.mapFilePath);
    if (ImGui::Button("Load .map file")) {
      editorState.currentMap = PLoadMapFromFile(editorState.mapFilePath);
      editorState.loadedMap = true;

      PBuildCollisionFromMap(&editorState.currentMap);
    }
    ImGui::SameLine();
    if (ImGui::Button("Create .map file")) {
      editorState.currentMap = PCreateNewMap(editorState.mapFilePath);
      editorState.loadedMap = true;

      PBuildCollisionFromMap(&editorState.currentMap);
    }
    ImGui::End();

    if (editorState.loadedMap) {
      ImGui::Begin("Map Editor Tools");

      ImGui::Text("Keybinds:");
      ImGui::Text("PageDown : Toggle place tile on click");

      if (ImGui::Button("Save level file")) {
        PSaveMapToFile(&editorState.currentMap, editorState.mapFilePath);
      }
      if (ImGui::Button("Place player spawn")) {
        editorState.placingPlayerSpawn = true;
        SDL_Log("Enabled player spawn point placing mode");
      }
      ImGui::Text("Viewport scale");
      ImGui::SameLine();

      if (ImGui::Button("-")) {
        editorState.editorScale -= 1;
      }
      ImGui::SameLine();
      if (ImGui::Button("+")) {
        editorState.editorScale += 1;
      }

      const char *tileNames[] = {"WALL", "FLOOR"};

      ImGui::Text("Tile placement");
      ImGui::InputInt("Tile count", &editorState.tilePlaceAmount);
      if (ImGui::BeginCombo("Selected tile", editorState.selectedTileName)) {
        for (int i = 0; i < IM_ARRAYSIZE(tileNames); i++) {
          bool isSelected = (editorState.selectedTileName == tileNames[i]);
          if (ImGui::Selectable(tileNames[i], isSelected)) {
            editorState.selectedTileName = tileNames[i];
            SDL_Log("Selected tile: %s", editorState.selectedTileName);
          }

          if (isSelected) {
            ImGui::SetItemDefaultFocus();
          }
        }

        ImGui::EndCombo();
      }

      ImGui::End();

      ImGui::Begin("Map Information Editor");
      ImGui::Text("WARNING! Changing map size and scale after you build your "
                  "map will break it!");
      ImGui::Text("NOTE! Map properties save on map save");
      ImGui::NewLine();
      ImGui::Separator();
      ImGui::NewLine();
      ImGui::InputText("Map name", &editorState.currentMap.mapName);
      ImGui::InputText("Map version", &editorState.currentMap.mapVersion);
      ImGui::InputText("Map author", &editorState.currentMap.mapAuthor);
      ImGui::InputInt("Map scale", &editorState.currentMap.mapScaleX);
      ImGui::InputInt("Map scale (Y)", &editorState.currentMap.mapScaleY);
      ImGui::InputInt("Map width", &editorState.currentMap.mapWidth);
      ImGui::InputInt("Map height", &editorState.currentMap.mapHeight);

      ImGui::End();
    }

    // Rendering
    ImGui::Render();

    SDL_RenderClear(state.renderer);

    SDL_RenderSetScale(state.renderer, editorState.editorScale,
                       editorState.editorScale);
    if (editorState.loadedMap) {
      PRenderMap(&editorState.currentMap, true);
    }

    SDL_RenderSetScale(state.renderer, io.DisplayFramebufferScale.x,
                       io.DisplayFramebufferScale.y);
    SDL_SetRenderDrawColor(state.renderer, (Uint8)(clear_color.x * 255),
                           (Uint8)(clear_color.y * 255),
                           (Uint8)(clear_color.z * 255),
                           (Uint8)(clear_color.w * 255));

    ImGui_ImplSDLRenderer_RenderDrawData(ImGui::GetDrawData());

    SDL_RenderPresent(state.renderer);
  }

  // Cleanup
  ImGui_ImplSDLRenderer_Shutdown();
  ImGui_ImplSDL2_Shutdown();
  ImGui::DestroyContext();

  SDL_DestroyRenderer(state.renderer);
  SDL_DestroyWindow(window);
  SDL_Quit();
}
